import logging
from functools import wraps
from inspect import signature
from typing import Callable, Dict, Tuple, Any

from aiohttp.client import ClientSession


LOGGER_NAME = 'webutils'
logger = logging.getLogger(LOGGER_NAME)


def _get_param_idx_and_default(function: Callable, param_name: str) -> Tuple[int, Any]:
    params = signature(function).parameters
    return list(params.keys()).index(param_name), params[param_name].default


# TODO: Figure out why PyCharm does not produce type hints to a function decorated in this manner,
#       when using the decorator without parentheses (e.g. @in_async_session instead of @in_async_session()
#       in this case), as soon as the `Callable` return type is added to that decorator's signature.
def in_async_session(_func: Callable = None, *,
                     session_kwargs: Dict[str, Any] = None, session_param_name: str = 'session'):
    """
    Useful decorator for any async function that uses the `aiohttp.ClientSession` to make requests.

    Using this decorator allows the decorated function to have an optional session parameter,
    without the need to ensure proper initialization and closing of a session within the function body itself.

    The wrapper has no effect, if the default value is passed to the session parameter during the function call,
    otherwise it initializes a session, passes it into the function and ensures that it is closed in the end.

    The wrapper distinguishes between positional and keyword arguments. This means that if the function is called by
    passing the default value as a positional argument, the temporary session is injected in its place as a positional
    argument also; if the default is passed as a keyword argument, the temporary session is injected as such instead.

    Args:
        _func:
            If this decorator is used *with any* arguments, this will always be the decorated function itself.
            This is a trick to allow the decorator to be used with as well as without arguments, i.e. in the form
            `@in_async_session` or `@in_async_session(...)`.
        session_kwargs (optional):
            If passed a dictionary, it will be unpacked and passed as keyword arguments into the `ClientSession`
            constructor, if and only if the decorator actually handles session initialization/closing,
            i.e. only when the function is called **without** passing a session object into it.
        session_param_name (optional):
            The name of the decorated function's parameter that should be passed the session object as an argument.
            In case the decorated function's session parameter is named anything other than "session", that name should
            be provided here.
    """
    if session_kwargs is None:
        session_kwargs = {}

    def decorator(function: Callable) -> Callable:
        # Using `functools.wraps` to preserve information about the actual function being decorated
        # More details: https://docs.python.org/3/library/functools.html#functools.wraps
        @wraps(function)
        async def wrapper(*args, **kwargs) -> Any:
            """The actual function wrapper that may perform the session initialization and closing."""
            temp_session = None
            session_param_idx, session_param_default = _get_param_idx_and_default(function, session_param_name)
            try:
                session = args[session_param_idx]
            except IndexError:
                # If we end up here, this means there was no positional session argument passed.
                session = kwargs.get(session_param_name)
                if session == session_param_default:
                    logger.debug(f"Injecting temporary `ClientSession` as keyword argument `{session_param_name}` "
                                 f"into `{function.__name__}`")
                    temp_session = ClientSession(**session_kwargs)
                    kwargs[session_param_name] = temp_session
            else:
                if session == session_param_default:
                    logger.debug(f"Injecting temporary `ClientSession` as positional argument {session_param_idx} "
                                 f"into `{function.__name__}`")
                    temp_session = ClientSession(**session_kwargs)
                    args = list(args)
                    args[session_param_idx] = temp_session
            try:
                return await function(*args, **kwargs)
            finally:
                if temp_session:
                    await temp_session.close()
                    logger.debug("Temporary `ClientSession` closed")
        return wrapper
    return decorator if _func is None else decorator(_func)
